#ifndef STR_H_INCLUDED
#define STR_H_INCLUDED
#pragma once

/* light-weight, safe, and efficient string classes that */ 
/* mix well with APIs that require plain LPTSTR strings. */ 

template <typename C>
class SimpleStringConstant
{
public:
  typedef C char_t;
  typedef SimpleStringConstant<char_t> same_t;

protected:
  const char_t *s, *e;

public:
  size_t size(void) const { return e-s; }
  bool empty(void) const { return e==s; }
  const char_t *begin(void) const { return s; }
  const char_t *end(void) const { return e; }

  SimpleStringConstant(const char_t *s_, size_t n)
    : s(s_), e(s_+n) { }
};

template <typename C>
class SimpleStringBuffer
{
/*
  
  TODO: write stl-style iterators so you can traverse the string
  in either direction with std::foreach.  (and for the above class)
  
*/
public:
  typedef C char_t;
  typedef SimpleStringBuffer<char_t> same_t;
  typedef SimpleStringConstant<char_t> const_t;

protected:
  char_t *s, *p;
  const char_t *e;

  bool append_(const char_t *s_, const char * const e_)
  {
    /* this could be rewritten to compile into tighter code */ 
    /* I think this is the clearest way to write it though. */ 
    while (s_<e_ && p<e)
      *p++ = *s_++;
    return s_==e_;
  }
  
  /* see below, there's a reason for this. */ 
  same_t &operator=(const same_t &);

public:
  size_t capacity(void) const { return e-s; }
  size_t size(void) const { return p-s; }
  size_t room(void) const { return e-p; }
  bool empty(void) const { return p==s; }
  bool full(void) const { return p==e; }
  char_t *begin(void) const { return s; }
  const char_t *end(void) const { return p; }
  const char_t *csafe(void)
  {
    if (!capacity()) return NULL;
    if (full()) --p;
    *p = 0;
    return begin();
  }

  void eraseall(void) { p=s; }

  bool copy(const same_t &r) { eraseall(); return append(r); }
  bool append(const same_t &r) { return append_(r.begin(), r.end()); }
  bool append(const const_t &r) { return append_(r.begin(), r.end()); }
  bool append(char_t r)
  {
    if (p==e) return false;
    *p++ = r; return true;
  }

  bool fit(void)
  {
    p = s;
    while (p<e && *p)
      ++p;
    if (p!=e)
      return true;
    p = s;
    return false;
  }

  same_t &operator<<(const same_t &r) { append(r); return *this; }
  same_t &operator<<(const const_t &r) { append(r); return *this; }
  same_t &operator<<(char_t r) { append(r); return *this; }

  char_t operator[](int i)
  {
    /* negative indexes start at the end */ 
    if (i<0)
      i += size();
    /* in theory, half of strings that would overflow int is unavailable through this */ 
    /* in practice, I doubt there are any real strings that long.                     */ 
    if (i<0 || static_cast<unsigned>(i)>=size())
      return 0;
    return begin()[i];
  }

  operator const_t(void) const
  { return const_t(begin(), size()); }

  SimpleStringBuffer(char_t *s_, size_t n)
    : s(s_), p(s_), e(s_+n) { }
  /* Note that the copy constructor does not actually */ 
  /* create a copy of the string.  It just copies the */ 
  /* pointers, effectively making a reference.        */ 
  /* That's why operator= is disabled, it would be    */ 
  /* misleading for it to perform either operation.   */ 
};

template <typename C, int N>
class StackStringBuffer : public SimpleStringBuffer<C>
{
public:
  enum { nBufferSize = N };

protected:
  char_t sBuffer_[nBufferSize];

public:
  StackStringBuffer(void) :  SimpleStringBuffer<C>(sBuffer_, nBufferSize) { }
  StackStringBuffer(const SimpleStringBuffer<C> &l)
    :  SimpleStringBuffer<C>(sBuffer_, nBufferSize)
  { copy(l); }
};

#ifndef TCHAR
#ifdef UNICODE
#define TCHAR wchar_t
#else
#define TCHAR char
#endif
#endif

#ifndef TEXT
#ifdef UNICODE
#define TEXT(N) L##N
#else
#define TEXT(N) N
#endif
#endif

template <typename C>
inline SimpleStringConstant<C> MakeStringConstant(const C *s, size_t n)
{ return SimpleStringConstant<C>(s, n); }

#define S(N) (MakeStringConstant(TEXT(N), sizeof(TEXT(N))-1))

/* This is a handy shortcut, but it means you can't make */ 
/* temporaries using the bufstr() constructor.  It seems */ 
/* like a fair trade off to me.                          */ 
#define bufstr(SIZE) StackStringBuffer<TCHAR, SIZE>
typedef SimpleStringBuffer<TCHAR> bufstr;
typedef SimpleStringConstant<TCHAR> str;
typedef const TCHAR *cstr;

#endif /* STR_H_INCLUDED */ 
